"""Contains tensorflow code for prediction models usage."""
import sys

from pathlib import Path

import tensorflow as tf
import tensorflow.contrib.slim as slim    # pylint: disable=E0401,W0611,R0402

from absl import logging


class Models:
    """Handles 'vgg-mix' and 'incep-mix' models predictions.

       Attributes:
          model_type: str, the type of graph to be used.
          models_root_path: str, path to model files.
    """

    def __init__(self, model_type: str, models_root_path: str):
        """Initializes session state."""
        model_path = Path(models_root_path).joinpath(model_type)
        model_graph = model_path.joinpath(model_type + ".ckpt.meta")
        self.session_config = tf.ConfigProto()
        self.session_config.gpu_options.allow_growth = True
        self.session = tf.Session(config=self.session_config)
        try:
            loader = tf.train.import_meta_graph(str(model_graph))
            loader.restore(self.session,
                           tf.train.latest_checkpoint(str(model_path)))
        except ValueError as err:
            logging.error("Missing models files - %s", err)
            sys.exit(1)
        self.image_tensor = self.session.graph.get_tensor_by_name(
            "Placeholder:0")
        self.isTest = self.session.graph.get_tensor_by_name("Placeholder_2:0")    # pylint: disable=C0103
        if model_type == "vgg-mix":
            self.scores = self.session.graph.get_tensor_by_name(
                "vgg_16/fc8/squeezed:0")
        elif model_type == "incep-mix":
            self.scores = self.session.graph.get_tensor_by_name(
                "InceptionV4/Logits/Logits/BiasAdd:0")
        else:
            logging.error("Unknown model type - %s", model_type)
            sys.exit(1)
        self.image_size = self.image_tensor.get_shape().as_list()[1:3]

    def get_size(self):
        """Finds out model's input tensor size.

        Returns:
            image_size: tuple, width and heigth of imput array.
        """
        return self.image_size

    def predict(self, inputs):
        """Runs prediction on input array.

        Returns:
            prediction: np.array, prediction results.
        """
        with self.session.as_default():
            prediction = self.session.run(self.scores, {
                self.image_tensor: inputs,
                self.isTest: True
            })
            return prediction



"""Command line interface for wsi-til library."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import signal
from pathlib import Path

from absl import app
from absl import flags
from absl import logging

from tily import handlers
from tily import prediction

FLAGS = flags.FLAGS
MODEL_URL = "https://bit.ly/wsi-til-zip"
WORKDIR = Path.cwd()

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
logging.set_verbosity(logging.ERROR)

flags.DEFINE_string("input_tiff", None, "Path to TIFF or SVS file.")
flags.DEFINE_string("model_type", "vgg-mix", "Prediction model to use.")
flags.DEFINE_string("model_path", "/tmp/wsi_til/models",
                    "Location of pre-trained models files.")
flags.DEFINE_string("model_url", MODEL_URL,
                    "Remote location of models archive.")
flags.DEFINE_string("output_path", str(WORKDIR),
                    "Location where to store csv with results.")
flags.DEFINE_float("mpp", 0.3444,
                   "Default MPP value to use if none found in TIFF metadata.")
flags.DEFINE_integer("page_index", 0, "Index of TIFF page to use as reference.")
flags.DEFINE_integer("patch_size", 100, "Size of slide patch to use as input.")
flags.DEFINE_integer("patch_magnification", 20,
                     "Magnification of patch to match with slide.")


def _handle_term(*args) -> None:
    """Termination handler."""
    raise KeyboardInterrupt()


def _must_get_input_tiffs() -> list:
    """Validates input TIFF files.

    Return:
        list, input tiffs files.
        exit(1), if file doesn't exists.
    """
    input_tiffs = []

    if FLAGS.input_tiff:
        absoulte_tiff_path = handlers.absolute_file_path(FLAGS.input_tiff)
        if not Path(absoulte_tiff_path).exists():
            print("Provided TIFF '%s' wasn't found.", FLAGS.input_tiff)
            sys.exit(1)
        input_tiffs.append(absoulte_tiff_path)

    if FLAGS.input_tiff is None:
        logging.warning("'input_tiff' flag wasn't set - "
                        "will look for TIFF and SVS files in current workdir.")
        logging.warning("Current workdir is '%s'", WORKDIR)

        def _folder_parser(extension: str) -> list:
            """Looks for files with extension in WORKDIR.

            Args:
                extension: str, name of the files extension.

            Returns:
                list, files with matching extension.
            """
            return list(WORKDIR.glob(f"*.{extension}"))

        local_tiffs = _folder_parser("tif") + _folder_parser("svs")
        for found_file in local_tiffs:
            input_tiffs.append(str(found_file))
    return input_tiffs


def _must_get_models():
    """Validates that models graph exists."""
    model_path = Path(handlers.absolute_file_path(FLAGS.model_path))
    model_type_path = model_path.joinpath(FLAGS.model_type)
    model_graph = model_type_path.joinpath(FLAGS.model_type + ".ckpt.meta")
    if not model_graph.exists():
        output_path = handlers.absolute_file_path(FLAGS.model_path)
        if not Path(output_path).exists():
            Path(output_path).mkdir(parents=True)
        logging.warning("Saving models graphs to '%s'", output_path)
        if not handlers.get_models(FLAGS.model_url, output_path):
            logging.error("Failed to download model files.")
            sys.exit(1)


def _render_startup_options(tiff_file: str) -> dict:
    """Renders startup flags into attributes.

    Args:
        tiff_file: str, absoulte file path to TIFF file.

    Returns:
        startup_options: dict, rendered startup attributes.
    """
    startup_options = {
        "input_tiff": tiff_file,
        "model_type": FLAGS.model_type,
        "model_path": handlers.absolute_file_path(FLAGS.model_path),
        "output_path": handlers.absolute_file_path(FLAGS.output_path),
        "mpp": FLAGS.mpp,
        "page_index": FLAGS.page_index,
        "patch_size": FLAGS.patch_size,
        "patch_magnification": FLAGS.patch_magnification,
    }
    return startup_options


def init(argv):
    """Initializes CLI interface."""
    del argv

    signal.signal(signal.SIGTERM, _handle_term)
    logging.info("Retrieving TIFF files.")
    input_tiffs = _must_get_input_tiffs()
    if len(input_tiffs) == 0:
        print("No TIFF files found - nothing to do.")
        sys.exit(1)

    logging.info("Validating model files.")
    _must_get_models()

    logging.info("Rendering startup options.")
    startup_options = [_render_startup_options(tiff) for tiff in input_tiffs]

    logging.info("Starting prediction runner.")
    for task in startup_options:
        try:
            prediction.WsiTil(task)
        except AttributeError as err:
            logging.error("Failed to process '%s' - '%s'", task["input_tiff"],
                          err)
        except KeyboardInterrupt:
            print("\nProcess terminated. Bye!\n")

def main():
    """PIP package entrypoint."""
    sys.exit(app.run(init))


if __name__ == "__main__":
    app.run(init)



"""Contains class for execution of prediction model."""
import csv
import time

from pathlib import Path

from alive_progress import alive_it    # pylint: disable=E0401

from tily import tiff
from tily import models

PREDICTION_IN_TENSOR_INDEX = 1


class WsiTil:
    """Prediction model for Tumor Infiltrating Lymphocytes slides.

    Attributes:
        input_tiff: str, absolute location of the TIFF file.
        model_type: str, the name of the prediction model.
        model_path: str, absolute location of the models files.
        output_path: str, absoulte location for the output files.
        mpp: float, the pixel mapping of the scanned image.
        page_index: int, index of TIFF page to use as reference.
        patch_size: int, size of patch to crop from original slide.
        patch_magnification: int, magnifacation of the cropped area.
    """

    def __init__(self, startup_options: dict) -> None:
        """Initializes class with required parameters."""

        self.tiff_file = startup_options["input_tiff"]
        self.width, self.height, self.crop_size = tiff.get_tiff_parameters(
            self.tiff_file, startup_options["mpp"],
            startup_options["page_index"],
            startup_options["patch_magnification"],
            startup_options["patch_size"])
        self.prediction_csv = self.csv_file_name(startup_options["output_path"])
        self.run_predictor(startup_options["model_type"],
                           startup_options["model_path"],
                           startup_options["page_index"])

    def csv_file_name(self, output_path: str) -> str:
        """Generates CSV file name."""
        tiff_name = str(Path(Path(self.tiff_file).name).stem)
        prediction_csv = str(
            Path(output_path).joinpath(tiff_name)
        ) + '_w_' + str(self.width // self.crop_size) + '_h_' + str(
            self.height // self.crop_size) + '_csz_' + str(
                self.crop_size) + time.strftime("_%Y%m%d_%H%M%S") + '.csv'
        return prediction_csv

    def csv_writer(self, data_row: tuple) -> None:
        """Writes rows to csv file.

        Args:
            data_row: tuple, containing data to append to csv.
        """
        headers = ('x', 'y', 'prediction', 'whiteness', 'blackness', 'redness')
        with open(self.prediction_csv, 'a',
                  encoding="utf-8") as prediction_file:
            writer = csv.writer(prediction_file)
            if not Path(self.prediction_csv).exists():
                writer.writerow(headers)
            writer.writerow(data_row)

    def run_predictor(self, model_type: str, model_path: str,
                      page_index: int) -> None:
        """Runs model prediction inference over TIFF crops."""

        page = tiff.get_tiff_page(self.tiff_file, page_index)
        model = models.Models(model_type, model_path)
        model_input_image_size = model.get_size()

        for x_index_crop_on_slide in alive_it(
                range(self.height // self.crop_size),
                title=f"Running {model_type} based prediction",
                theme="scuba"):
            for y_index_crop_on_slide in range(self.width // self.crop_size):
                x_top_left = x_index_crop_on_slide * self.crop_size
                y_top_left = y_index_crop_on_slide * self.crop_size
                crop_array = tiff.get_crop_array(page, x_top_left, y_top_left,
                                                 self.crop_size)
                resized_array = tiff.resize_image_array(crop_array,
                                                        model_input_image_size)

                prediction = model.predict(
                    tiff.normalize_image_array(resized_array))
                prediction_sigmoid = tiff.get_sigmoid(
                    prediction[:, PREDICTION_IN_TENSOR_INDEX])
                self.csv_writer(
                    (x_index_crop_on_slide, y_index_crop_on_slide,
                     prediction_sigmoid, tiff.get_whiteness(resized_array),
                     tiff.get_blackness(resized_array),
                     tiff.get_redness(resized_array)))



"""TIFF files processing functions."""
import xml.etree.ElementTree as ET

from typing import Any, Optional, Tuple

import tifffile
import numpy as np

from absl import logging
from PIL import Image

PAGE_DECODE_TILE_INDEX = 0
SAMPLE_DIMENSION_INDEX = 0
RED_CHANNEL_INDEX = 0
GREEN_CHANNEL_INDEX = 1
BLUE_CHANNEL_INDEX = 2


def get_tiff_properties(page: Any) -> dict:
    """Converts TIFF page properties to dictionary.

    Args:
        page: , single page of the input file.

    Returns:
        properties: dict, tag.name:tag.value parameters for the page.
    """
    properties = {tag.name: tag.value for tag in page.tags}
    return properties


def get_mpp_value(tiff_properties: dict, default_mpp: float) -> float:
    """Parses metadata properties for MPP value.

    Args:
        tiff_properties: dict, TIFF page parameters.
        default_mpp: float, default MPP value if not provided in metadata.

    Returns:
        float, TIFF page MPP value.
    """
    xml = tiff_properties.get('OlympusSIS')
    tiff_values = tiff_properties.get('ImageDescription')
    if tiff_values:
        if xml:
            tiff_values_xml = ET.fromstring(tiff_values)
            for pixels in tiff_values_xml.findall(
                    ".//{http://www.openmicroscopy.org/Schemas/OME/2015-01}Pixels"
            ):
                physical_size_x_value = pixels.attrib['PhysicalSizeX']
                return round(float(physical_size_x_value), 4)
        tiff_values_string = tiff_values.split('|')
        for value in tiff_values_string:
            if 'mpp' in value.lower():
                mpp_value = value.split("=")
                return round(float(mpp_value[-1].strip(" ")), 4)
    logging.warning(
        "No MPP value was found in metadata. "
        "Falling back to default: %s", default_mpp)
    return default_mpp


def get_width_value(tiff_properties: dict) -> int:
    """Extracts width(X) value from TIFF metadata.

    Args:
        tiff_properties: dict, TIFF page parameters.

    Returns:
        uint, TIFF page MPP value.
    """
    return int(tiff_properties.get('ImageWidth'))


def get_height_value(tiff_properties: dict) -> int:
    """Extracts height(Y) value from TIFF metadata.

    Args:
        tiff_properties: dict, TIFF page parameters.

    Returns:
        uint, TIFF page MPP value.
    """
    return int(tiff_properties.get('ImageLength'))


def get_crop_size(mpp_value: float, patch_magnification: int,
                  patch_size: int) -> int:
    """Calculates slide crop size based on MPP, patch size and magnification.

    Args:
        mpp_value: float, TIFF MPP value.
        patch_magnification: int, patch magnification value.
        patch_size: int, patch size value.

    Returns:
        uint, crop size value.
    """
    crop_size = (patch_size * (10.0 / mpp_value)) / patch_magnification
    return int(crop_size)


def get_tiff_parameters(file_path: str, default_mpp: float, page_index: int,
                        patch_magnification: int,
                        patch_size: int) -> Tuple[int, int, int]:
    """Calculates TIFF file parameters.

    Args:
        file_path: str, absolute path to TIFF file.
        default_mpp: float, fallback MPP value.
        page_index: int, index of TIFF page.
        patch_magnification: int, slide magnification.
        patch_size: int, slide patch size.

    Returns:
        width: int, image width.
        height: int, image height.
        crop: int, crop size.
    """
    with tifffile.TiffFile(file_path) as tif:
        properties = get_tiff_properties(tif.pages[page_index])
        mpp = get_mpp_value(properties, default_mpp)
        width = get_width_value(properties)
        height = get_height_value(properties)
        crop_size = get_crop_size(mpp, patch_magnification, patch_size)
        return int(width), int(height), int(crop_size)


def get_tiff_page(file_path: str, page_index: int):
    """Returns specified page

    Args:
        file_path: str, absolute path to TIFF file.
        page_index: int, index of TIFF page.

    Returns:
        page: Any, the TIFF page.
    """
    # with tifffile.TiffFile(file_path) as tif:
    #     page = tif.pages[page_index]
    #     return page
    return tifffile.TiffFile(file_path).pages[page_index]


def validate_crop_area(page: Any, x_top_left: int, y_top_left: int,
                       crop_size: int) -> Optional[list]:
    """Validates a crop area upon the TIFF page.

    Args:
        page: TiffPage, TIFF page from which the crop must be extracted.
        x_top_left: int, x coordinate of the top left corner of the desired crop.
        y_top_left: int, y coordinate of the top left corner of the desired crop.
        crop_size: int, desired crop size.

    Returns:
        err: str, error message if any.
    """
    errors = []
    if not page.is_tiled:
        err = "Input page must be tiled."
        logging.error(err)
        errors.append(err)

    if crop_size < 1:
        err = "Crop size is %s - must be strictly positive."
        logging.error(err, crop_size)
        errors.append(err)

    if x_top_left < 0:
        err = "Coordinates %s of X top left is out of image bounds."
        logging.error(err, x_top_left)
        errors.append(err)

    if y_top_left < 0:
        err = "Coordinates %s of Y top left is out of image bounds."
        logging.error(err, y_top_left)
        errors.append(err)

    if x_top_left + crop_size >= page.imagelength:
        err = "Coordinates %s of X bottom right is bigger than image X limit %s."
        logging.error(err, x_top_left + crop_size, page.imagelength)
        errors.append(err)

    if y_top_left + crop_size >= page.imagewidth:
        err = "Coordinates %s of Y bottom right is bigger than image Y limit %s."
        logging.error(err, y_top_left + crop_size, page.imagewidth)
        errors.append(err)

    if len(errors) > 0:
        return errors
    return None


def get_crop_array(    # pylint: disable=too-many-locals
        page: Any, x_top_left: int, y_top_left: int, crop_size: int) -> Any:
    """Extracts a crop from the TIFF page.

    Args:
        page: TiffPage, TIFF page from which the crop must be extracted.
        x_top_left: int, x coordinate of the top left corner of the desired crop.
        y_top_left: int, y coordinate of the top left corner of the desired crop.
        crop_size: int, desired crop size.

    Returns:
        np.ndarray, cropped slide area.
    """

    if validate_crop_area(page, x_top_left, y_top_left, crop_size):
        logging.error("Crop area is not valid.")
        return None

    # Calculate 2d limits (top left, bottom right) of tiles containing crop.
    tile_x_top_left = x_top_left // page.tilewidth
    tile_y_top_left = y_top_left // page.tilewidth
    tile_x_bottom_right, tile_y_bottom_right = np.ceil([
        (x_top_left + crop_size) / page.tilewidth,
        (y_top_left + crop_size) / page.tilewidth
    ]).astype(int)

    # Calculate the key for 1d/2d tile index conversion.
    tile_per_width = int(np.ceil(page.imagewidth / page.tilewidth))

    # Create an empty array to hold tiles containing crop.
    tiles_with_crop_arr = np.empty(
        ((tile_x_bottom_right - tile_x_top_left) * page.tilewidth,
         (tile_y_bottom_right - tile_y_top_left) * page.tilewidth,
         page.samplesperpixel),
        dtype=page.dtype)

    jpeg_tables = page.tags.get('JPEGTables', None)
    if jpeg_tables is not None:
        jpeg_tables = jpeg_tables.value

    for tile_x_2d in range(tile_x_top_left, tile_x_bottom_right):
        for tile_y_2d in range(tile_y_top_left, tile_y_bottom_right):
            # Convert tile 2d index back to 1d.
            tile_index_1d = int(tile_x_2d * tile_per_width + tile_y_2d)
            # Get tile array.
            page.parent.filehandle.seek(page.dataoffsets[tile_index_1d])
            data = page.parent.filehandle.read(
                page.databytecounts[tile_index_1d])
            # Get top left corner of tile array on tiles_with_crop_arr.
            im_x = (tile_x_2d - tile_x_top_left) * page.tilewidth
            im_y = (tile_y_2d - tile_y_top_left) * page.tilewidth
            # write tile array without sample dimension into tiles_with_crop_arr.
            tiles_with_crop_arr[
                im_x:im_x + page.tilewidth,
                im_y:im_y + page.tilewidth, :] = page.decode(
                    data, tile_index_1d,
                    jpeg_tables)[PAGE_DECODE_TILE_INDEX].squeeze(
                        axis=SAMPLE_DIMENSION_INDEX)
    # Get top left corner of crop on tiles_with_crop_arr.
    im_x_top_left = x_top_left - tile_x_top_left * page.tilewidth
    im_y_top_left = y_top_left - tile_y_top_left * page.tilewidth

    return tiles_with_crop_arr[im_x_top_left:im_x_top_left + crop_size,
                               im_y_top_left:im_y_top_left + crop_size, :]


def resize_image_array(crop_array: np.ndarray,
                       target_size: tuple) -> np.ndarray:
    """Resizes the crop array.

    Args:
        crop_array: np.ndarray, crop 3d RGB array to resize.
        target_size: tuple, width and height of desired image array.

    Returns:
        np.ndarray, 3d RGB array of model-requ
        ired width and height.
    """
    image_resized = Image.fromarray(
        crop_array.astype('uint8')).resize(target_size)
    return np.asarray(image_resized)


def get_whiteness(crop_array: np.ndarray) -> float:
    """Measures the whiteness of the input image array.

    Args:
        crop_array: np.ndarray, crop 3d RGB array of uint8 of range 0..255.

    Returns:
        whiteness: float, mean standard deviation across 3 color channels.
    """
    whiteness = (np.std(crop_array[:, :, RED_CHANNEL_INDEX]) +
                 np.std(crop_array[:, :, GREEN_CHANNEL_INDEX]) +
                 np.std(crop_array[:, :, BLUE_CHANNEL_INDEX])) / 3.0
    return float(whiteness)


def get_blackness(crop_array: np.ndarray) -> float:
    """Measures the blackness of the input image array.

    Args:
        crop_array: np.ndarray, crop 3d RGB array of uint8 of range 0..255.

    Returns:
        blackness: float, mean intensity across 3 color channels.
    """
    blackness = np.mean(crop_array)
    return float(blackness)


def get_redness(crop_array: np.ndarray) -> float:
    """Measures the redness of the input image array.

    Args:
        crop_array: np.ndarray, crop 3d RGB array of uint8 of range 0..255.

    Returns:
        redness: float, mean of red-weighted intensity across 3 color channels.
    """
    redness = np.mean((crop_array[:, :, RED_CHANNEL_INDEX] >= 190) *
                      (crop_array[:, :, GREEN_CHANNEL_INDEX] < 100) *
                      (crop_array[:, :, BLUE_CHANNEL_INDEX] <= 100))
    return float(redness)


def normalize_image_array(crop_array: np.ndarray) -> np.ndarray:
    """Normalizes the array for model input.

    Args:
        crop_array: np.ndarray, image array of uint8 of range 0..255.

    Returns:
        normalized_array: np.ndarray, normalized 4d float array of range -1..1.
    """
    normalized_array = np.empty(crop_array.shape)
    np.clip(crop_array, 0, 255, normalized_array)
    normalized_array /= 255
    normalized_array -= 0.5
    normalized_array *= 2
    return np.expand_dims(normalized_array, axis=SAMPLE_DIMENSION_INDEX)


def get_sigmoid(prediction: float) -> float:
    """Computes sigmoid value of prediction.

    Args:
        prediction: float, model inference output.

    Returns:
        sig: float, sigmoid - original distributed values pushed to 0..1.
    """
    sigmoid = 1 / (1 + np.exp(-prediction))
    return float(sigmoid)



"""Setup tily."""

from pathlib import Path
from setuptools import find_packages, setup

HERE = Path(__file__).parent
README = HERE.joinpath("README.md")

with open(README) as f:
    readme = f.read()

setup(
    name="tily",
    version="0.0.2",
    python_requires=">=3.7.4, <=3.7.12",
    description="Tumor-infiltrating lymphocytes prediction based on Whole Slide Imaging.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/kandabarau/wsi-til",
    author="Sergey Kandabarau",
    author_email="kandabarau@gmail.com",
    license="BSD",
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(include=["tily"], exclude=["tests*"]),
    install_requires=[
        "about-time==3.1.1",
        "absl-py==1.0.0",
        "alive-progress==2.1.0",
        "astor==0.8.1",
        "attrs==21.2.0",
        "certifi==2021.10.8",
        "charset-normalizer==2.0.9",
        "gast==0.2.2",
        "google-pasta==0.2.0",
        "grapheme==0.6.0",
        "grpcio==1.42.0",
        "h5py==2.10.0",
        "idna==3.3",
        "imagecodecs==2021.11.20",
        "importlib-metadata==4.8.2",
        "iniconfig==1.1.1",
        "keras-applications==1.0.8",
        "keras-preprocessing==1.1.2",
        "markdown==3.3.6",
        "numpy==1.17.3",
        "opt-einsum==3.3.0",
        "packaging==21.3",
        "pillow==8.4.0",
        "pluggy==1.0.0",
        "protobuf==3.19.1",
        "py==1.11.0",
        "pyparsing==3.0.6",
        "pytest==6.2.5",
        "requests==2.26.0",
        "six==1.16.0",
        "tensorboard==1.15.0",
        "tensorflow==1.15.5",
        "tensorflow-estimator==1.15.1",
        "termcolor==1.1.0",
        "tifffile==2021.11.2",
        "toml==0.10.2",
        "typing-extensions==4.0.1",
        "urllib3==1.26.7",
        "werkzeug==2.0.2",
        "wheel==0.37.0",
        "wrapt==1.13.3",
        "zipp==3.6.0",
        ],
    entry_points={
        "console_scripts": [
            "tily=tily.__main__:main",
        ]
    },
)



"Handlers are some helper functions that are required for prediction module."
import shutil
import tempfile
import zipfile

from pathlib import Path
from typing import Optional

import requests

from absl import logging
from alive_progress import alive_it    # pylint: disable=E0401

CHUNK_SIZE = 1024
SUPPORTED_MODELS = ["vgg-mix", "incep-mix"]


def get_models(url: str, models_path: str) -> bool:
    """Prepares models files to be used for prediction.

    Args:
        url: str, remote location of models archive.
        models_path: str, path to directory with TF models.

    Returns:
        result: bool, True if sucessfull, False if not.
    """
    result = True
    temp_dir = tempfile.mkdtemp()
    logging.info("Temporary folder '%s' created", temp_dir)

    logging.info("Downloading archive from %s", url)
    model_zip_file = download_file(url, "models.zip", temp_dir)
    if model_zip_file is None:
        result = False

    zip_dir = _unpack_model(model_zip_file)
    if zip_dir is None:
        result = False
    logging.info("'%s' was extracted to '%s'", model_zip_file, zip_dir)

    model_path = _sort_model(zip_dir, models_path)
    logging.info("The model was sorted.")

    _add_checkpoints(model_path)
    logging.info("Checkpoints added.")

    logging.info("Removing temporary folder '%s'", temp_dir)
    shutil.rmtree(temp_dir)
    return result


def download_file(url: str, file_name: str, output_path: str) -> Optional[str]:
    """Downloads file from remote url.

    Args:
        url: str, link to remote location.
        file_name: str, expected file name with extension.
        output_path: str, location to save.

    Returns:
        file_path: Optional[str], local path to downloaded file.
    """
    file_path = Path(output_path).joinpath(file_name)
    if Path(absolute_file_path(file_path)).exists():
        return str(file_path)
    try:
        req = requests.get(url, allow_redirects=True, stream=True)
        file_type = f"application/{Path(file_name).suffix[1:]}"
        content_type = req.headers["Content-Type"]
        if content_type != file_type:
            logging.warning("Expecteda remote to be '%s' - got '%s'.",
                            file_type, content_type)
    except requests.exceptions.MissingSchema as err:
        logging.error("Please check the url - %s", err)
        return None

    if not Path(absolute_file_path(str(output_path))).is_dir():
        logging.info("Creating output folder '%s'.", output_path)
        Path(absolute_file_path(output_path)).mkdir(parents=True, exist_ok=True)
    with open(file_path, "wb") as file:
        for chunk in alive_it(req.iter_content(chunk_size=CHUNK_SIZE),
                              title=f"Downloading {file_name}",
                              theme="scuba"):
            file.write(chunk)
    return str(file_path)


def absolute_file_path(file_path: str) -> str:
    """Takes relative file path and returns absolute.

    Args:
        file_path: any form of path.

    Returns:
        str, resolved absolute file path.
    """
    return str(Path(file_path).resolve())


def _unpack_model(archive_path: str) -> Optional[str]:
    """Unpacks models archive.

    Args:
        archive_path: str, location of the zip archive.

    Returns:
        zip_dir: str, temporary local extraction path.
    """
    try:
        zip_dir = Path(archive_path).parent
        logging.info("Model archive will be extracted to '%s'", zip_dir)
        with zipfile.ZipFile(archive_path, "r") as zip_ref:
            zip_ref.extractall(zip_dir)
        return str(zip_dir)
    except zipfile.BadZipFile as err:
        logging.error("'%s' is not zip archive - %s", archive_path, err)
        return None
    except FileNotFoundError as err:
        logging.error("'%s' doesn't exists - %s", archive_path, err)
        return None


def _sort_model(input_model_path: str, output_path: str) -> str:
    """Sorts necessary files and saves them to final location.

    Args:
        input_model_path: str, unsorted model path location.
        output_path: str, destination location.

    Returns:
        output_model_path: str, absolute path to saved models.
    """
    source = Path(input_model_path).glob("*ckpt*")
    model_files = [f for f in source if f.is_file()]

    model_types = set()
    for file in model_files:
        model_types.add(Path(file.stem).stem)

    output_model_path = Path(output_path)
    for model in model_types:
        model_path = output_model_path.joinpath(model)
        logging.info("Creating folder for model '%s' as '%s'", model,
                     model_path)
        model_path.mkdir(parents=True, exist_ok=True)
        for file in model_files:
            if model in file.name:
                model_file = Path(model_path).joinpath(file.name)
                logging.info("Saving '%s' as '%s'", file.name, model_file)
                shutil.move(str(Path(file)), model_file)
    return str(output_model_path)


def _add_checkpoints(model_path: str) -> None:
    """Creates checkpoint files for each model.

    Args:
        model_path: str, location of sorted model files.
    """
    models = [d for d in Path(model_path).iterdir() if d.is_dir()]
    for model in models:
        if str(model.stem) in SUPPORTED_MODELS:
            checkpoint = model.joinpath('checkpoint')
            logging.info("Saving checkpoint for %s as %s", model, checkpoint)
            with open(checkpoint, "w", encoding="utf8") as file:
                file.write(f"model_checkpoint_path: \"{model.stem}.ckpt\"\n")
                file.write(
                    f"all_model_checkpoint_paths: \"{model.stem}.ckpt\"\n")
